--
-- GuildArtisan
-- 
-- Copyright (c) 2021 Dana Basken
--

local VERSION="9.0.5.1"

local REALM = GetRealmName()
local NAME_REALM = "%s-%s"
local PLAYER_NAME = UnitName("player");
local PLAYER_NAME_REALM = NAME_REALM:format(PLAYER_NAME, REALM)

GuildArtisan = LibStub("AceAddon-3.0"):NewAddon(GuildArtisan_Config.ADDON_NAME, "AceConsole-3.0", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0")

print(PrefixString(GuildArtisan_Config.ADDON_NAME, ("v%s"):format(VERSION), GuildArtisan_Config.PREFIX_COLOR))

local OPTIONS = buildOptions(GuildArtisan_Config.ADDON_NAME, GuildArtisan)

local AceGUI = LibStub("AceGUI-3.0") -- needed for AceConfigDialog
local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

function GuildArtisan:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("GuildArtisanDB", GuildArtisan_Config.DEFAULTS, true)
  self.minimapIcon = MinimapIcon:Create(("%sMinimap"):format(GuildArtisan_Config.ADDON_NAME), "Interface/Icons/INV_Chest_Cloth_17", self.db, function() self.window:Toggle() end)

  self.messageQueue = {}
  self.cache = GuildArtisanCache:Create(self.db)
  self.window = GuildArtisanWindow:Create(self.db, self.cache)

  self.cacheUpdateThrottle = Throttle:Create(1)
  self.loadedEventThrottle = Throttle:Create(1)

  self.guildMembershipState = StateMachine:Create("GUILD_MEMBERSHIP")
  self.activeTradeSkill = StateMachine:Create("ACTIVE_TRADE_SKILL")

  LibStub("AceConfig-3.0"):RegisterOptionsTable(GuildArtisan_Config.ADDON_NAME, OPTIONS)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(GuildArtisan_Config.ADDON_NAME)

  GuildArtisan:RegisterMessage("STATE_CHANGE_GUILD_MEMBERSHIP")
  GuildArtisan:RegisterMessage("STATE_CHANGE_ACTIVE_TRADE_SKILL")

  GuildArtisan:RegisterEvent("PLAYER_GUILD_UPDATE")
  GuildArtisan:RegisterEvent("TRADE_SKILL_LIST_UPDATE")
  GuildArtisan:RegisterEvent("TRADE_SKILL_CLOSE")

  GuildArtisan:RegisterChatCommand(GuildArtisan_Config.CHAT_COMMAND_NAME, "HandleChatCommand")

  GuildArtisan:RegisterComm(GuildArtisan_Config.ADDON_NAME)

  AceEvent.RegisterMessage(self, "GUILD_ARTISAN_CACHE_UPDATE")
  AceEvent.RegisterMessage(self, "GUILD_ARTISAN_CACHE_UPDATED")

  AceTimer.ScheduleRepeatingTimer(self, "PeriodicTasks", 30)
  AceTimer.ScheduleRepeatingTimer(self, "ProcessMessageQueue", 1)
end

function GuildArtisan:OnEnable()
  self:UpdateGuildState()
end

function GuildArtisan:STATE_CHANGE_GUILD_MEMBERSHIP(event, state)
  -- self:print(("%s: %s"):format(event, state))
  if state == "member" then
    self:SendAddonEvent("loaded", VERSION)
  else
    self.cache:SetGuildName(nil)
    self.window:SetGuildName(nil)
  end
end

function GuildArtisan:STATE_CHANGE_ACTIVE_TRADE_SKILL(event, state)
  -- self:print(("%s: %s"):format(event, state))
  if self.activeTradeSkill.state == "" then
    self.currentRecipeIds = nil
  end
end

function GuildArtisan:PLAYER_GUILD_UPDATE(...)
  self:UpdateGuildState()
end

function GuildArtisan:TRADE_SKILL_LIST_UPDATE(...)
  if self.guildMembershipState.state == "member" then
    if not C_TradeSkillUI.IsTradeSkillLinked() then
      self:UpdateCurrentTradeSkill()
    else
      self.activeTradeSkill:SetState("")
    end
  end
end

function GuildArtisan:TRADE_SKILL_CLOSE(...)
  self.activeTradeSkill:SetState("")
end

function GuildArtisan:GetSubCategories(parent)
  local categories = {}
  for _, categoryID in pairs({C_TradeSkillUI.GetSubCategories(parent.categoryID)}) do
    local category = C_TradeSkillUI.GetCategoryInfo(categoryID)
    table.insert(categories, category)
    for _, child in pairs(self:GetSubCategories(category)) do
      table.insert(categories, child)
    end
  end
  return categories
end

function GuildArtisan:GetLatestTradeSkillCategories()
  local categoryName = GuildArtisan_Config.EXPANSIONS[self.db.char.expansion].skills[self.activeTradeSkill.state]
  local categories = {C_TradeSkillUI.GetCategories()}
  for _, categoryID in pairs(categories) do
    local category = C_TradeSkillUI.GetCategoryInfo(categoryID)
    if category.name == categoryName then
      return self:GetSubCategories(category)
    end
  end
  return {}
end

function GuildArtisan:GetCurrentTradeSkill()
  local link, id = C_TradeSkillUI.GetTradeSkillListLink()
  local name, rank, icon, castTime, minRange, maxRange = GetSpellInfo(id)
  return name
end

function GuildArtisan:ShouldIncludeRecipe(recipe)
  if not recipe.learned then return false end
  if not recipe.craftable then return false end
  if self.db.char.specialRecipesOnly then
    if self.db.char.expansion == "shadowlands" then
      if recipe.unlockedRecipeLevel == nil then return false end -- all done
      if recipe.unlockedRecipeLevel == 1 and recipe.currentRecipeExperience == 0 then return false end -- no progress at all
    end
  end
  return true
end

function GuildArtisan:UpdateCurrentTradeSkill()
  self.activeTradeSkill:SetState(self:GetCurrentTradeSkill())
  local recipeIds = self:GetCurrentRecipeIds()
  for index, recipeId in pairs(recipeIds) do
    local recipe = C_TradeSkillUI.GetRecipeInfo(recipeId)
    self.cache:SetRecipe(PLAYER_NAME_REALM, self.activeTradeSkill.state, recipe.icon, recipe.name, recipe.unlockedRecipeLevel, recipe.currentRecipeExperience, recipe.nextLevelRecipeExperience)
  end
end

function GuildArtisan:GetCurrentRecipeIds()
  if not self.currentRecipeIds then
    self.currentRecipeIds = {}
    local recipeIds = C_TradeSkillUI.GetAllRecipeIDs()
    for _, category in pairs(self:GetLatestTradeSkillCategories()) do
      if category.enabled then
        for _, recipeId in pairs(recipeIds) do
          local recipe = C_TradeSkillUI.GetRecipeInfo(recipeId)
          if recipe.categoryID == category.categoryID and self:ShouldIncludeRecipe(recipe) then
            table.insert(self.currentRecipeIds, recipeId)
          end
        end
      end
    end
  end
  return self.currentRecipeIds
end

function GuildArtisan:ArrayContains(array, value)
  if #array > 0 then
    for i = 1, #array do
      if array[i] == value then return true end
    end
  end
  return false
end

function GuildArtisan:SendAddonEvent(event, data)
  if not data then data = "none" end
  self:SendAddonMessage(("%s:%s"):format(event, data))
end

function GuildArtisan:SendAddonMessage(message)
  if not self:ArrayContains(self.messageQueue, message) then
    table.insert(self.messageQueue, message)
  end
end

function GuildArtisan:ProcessMessageQueue()
  if #self.messageQueue > 0 then
    local message = table.remove(self.messageQueue)
    GuildArtisan:SendCommMessage(GuildArtisan_Config.ADDON_NAME, message, "GUILD")
  end
end

function GuildArtisan:UpdateGuildState()
  if (IsInGuild()) then
    local name, _, _ = GetGuildInfo("player")
    self.cache:SetGuildName(name)
    self.window:SetGuildName(name)
    self.guildMembershipState:SetState("member")
  else
    self.guildMembershipState:SetState("non-member")
  end
end

function GuildArtisan:PeriodicTasks()
end

function GuildArtisan:GetOption(info)
  local property = info.option.arg
  return self.db.char[property]
end

function GuildArtisan:SetOption(info, value)
  local property = info.option.arg
  self.db.char[property] = value
  self:AddonPrint("set '%s' to %s", info.option.arg, ObjectToString(value))
end

function GuildArtisan:GetMimimapVisible(info)
  return not self.db.char.minimap.hide
end

function GuildArtisan:SetMimimapVisible(info, value)
  self.db.char.minimap.hide = not value
  self.minimapIcon:UpdateVisibility()
  self:AddonPrint("set 'minimap' to %s", ObjectToString(value))
end

function GuildArtisan:AddonPrint(format, ...)
  print(PrefixString(GuildArtisan_Config.ADDON_NAME, string.format(format, ...), GuildArtisan_Config.PREFIX_COLOR))
end

function GuildArtisan:HandleChatCommand(input)
  if not input or input:trim() == "" then
    self:AddonPrint("unknown command: '%s', try '/%s help'", input, GuildArtisan_Config.CHAT_COMMAND_NAME)
  else
    if input == "help" then
      self:AddonPrint("%s", "Commands:")
      self:AddonPrint("  %s", "/ga clear - Clear cache.")
      self:AddonPrint("  %s", "/ga config - Open addon config panel.")
      self:AddonPrint("  %s", "/ga lock - Lock the window position.")
      self:AddonPrint("  %s", "/ga unlock - Unlock the window position.")
      self:AddonPrint("  %s", "/ga toggle - Toggle the window visibility.")
      self:AddonPrint("  %s", "/ga minimap - Toggle the minimap icon visibility.")
    else
      if input == "clear" then
        self.cache:Clear()
      elseif input == "toggle" then
        self.window:Toggle()
      elseif input == "minimap" then
        self.minimapIcon:Toggle()
      elseif input == "lock" then
        self.window:Lock()
        self:AddonPrint("Window locked.")
      elseif input == "unlock" then
        self.window:Unlock()
        self:AddonPrint("Window unlocked.")
      elseif input == "debug" then
        self:SetOption({option = {arg = "debug"}}, not self.db.char.debug)
      elseif input == "cache" then
        self:print(ObjectToString(self.cache:GetCache()))
      elseif input == "send" then
        self:SendCachedPatterns()
      elseif input == "options" or input == "config" then
        LibStub("AceConfigDialog-3.0"):Open("GuildArtisan")
      else
        self:AddonPrint("unknown command: '%s', try '/%s help'", input, GuildArtisan_Config.CHAT_COMMAND_NAME)
      end
    end
  end
end

function GuildArtisan:print(...)
  if self.db.char.debug then
    print(...)
  end
end

function GuildArtisan:MakeNumericVersion(version)
  local _, _, major, minor, patch, build = strfind(version, "(%d+)%.(%d+)%.(%d+)%.(%d+)")
  return (major * 1000000) + (minor * 10000) + (patch * 100) + build
end

function GuildArtisan:IsNewerVersion(version)
  local currentVersion = self:MakeNumericVersion(VERSION)
  local comparisonVersion = self:MakeNumericVersion(version)
  if comparisonVersion > currentVersion then
    return true
  else
    return false
  end
end

function GuildArtisan:UpdatePatternCache(data)
  local player, version, profession, recipeId, recipeName, level, currentXp, requiredXp = strsplit(":", data)
  self.cache:Update(player, profession, tonumber(recipeId), recipeName, tonumber(version), tonumber(level), tonumber(currentXp), tonumber(requiredXp))
end

function GuildArtisan:SendCachedPatterns()
  local messages = self.cache:ToMessages()
  for i = 1, #messages do
    self:SendAddonEvent("pattern", messages[i])
  end
end

function GuildArtisan:HandleLoadedEvent()
  self.loadedEventThrottle:SetWindowSeconds(math.random(1, 5))
  self.loadedEventThrottle:Do(function()
    self:SendCachedPatterns()
  end)
end

function GuildArtisan:GUILD_ARTISAN_CACHE_UPDATE(event, player, profession, recipeName)
  local message = self.cache:ToMessage(player, profession, recipeName)
  self:SendAddonEvent("pattern", message)
end

function GuildArtisan:GUILD_ARTISAN_CACHE_UPDATED(...)
  self.cacheUpdateThrottle:Do(function()
    self.window:RefreshUI()
  end)
end

function GuildArtisan:OnCommReceived(prefix, message, distribution, sender)
  local event, data = strsplit(":", message, 2)
  self:print("MESSAGE", prefix, event, data, sender)
  if sender ~= PLAYER_NAME and sender ~= PLAYER_NAME_REALM then
    if event == "loaded" then
      self:HandleLoadedEvent()
      if self:IsNewerVersion(data) then
        self:AddonPrint("new version available: v%s", data)
      end
    elseif event == "pattern" then
      self:UpdatePatternCache(data)      
    else
      self:print(("unknown message: %s"):format(message))
    end
  end
end
