--
-- GuildArtisanCache
--
-- Copyright (c) 2021 Dana Basken
--

local AceEvent = LibStub("AceEvent-3.0")

GuildArtisanCache = {}
GuildArtisanCache.__index = GuildArtisanCache

function GuildArtisanCache:Create(db)
  local instance = {}
  setmetatable(instance, GuildArtisanCache)
  instance:Construct(db)
  return instance
end

function GuildArtisanCache:Construct(db)
  self.db = db
  self.updated = {}
  self.readyState = StateMachine:Create("CACHE_READY")
  AceEvent.RegisterMessage(self, "STATE_CHANGE_CACHE_READY")
end

function GuildArtisanCache:SetGuildName(name)
  self.guildName = name
  if name then
    self.readyState:SetState("ready")
  else
    self.readyState:SetState("not-ready")
  end
end

function GuildArtisanCache:STATE_CHANGE_CACHE_READY(event, state)
  -- print(("%s: %s"):format(event, state))
  AceEvent:SendMessage("GUILD_ARTISAN_CACHE_UPDATED")
end

function GuildArtisanCache:IsReady()
  if self.readyState.state == "ready" then return true else return false end
end

function GuildArtisanCache:GetCache()
  if not self.db.profile.cache then self.db.profile.cache = {} end
  if not self.db.profile.cache[self.guildName] then self.db.profile.cache[self.guildName] = {} end
  return self.db.profile.cache[self.guildName]
end

function GuildArtisanCache:Clear()
  self.db.profile.cache[self.guildName] = {}
  AceEvent:SendMessage("GUILD_ARTISAN_CACHE_UPDATED")
end

function GuildArtisanCache:EnsureEntryExists(player, profession, recipeName)
  local cache = self:GetCache()
  if not cache[player] then self:InitializePlayer(player) end
  if not cache[player][profession] then self:InitializeProfession(player, profession) end
  if not cache[player][profession][recipeName] then self:InitializeRecipe(player, profession, recipeName) end
end

function GuildArtisanCache:Get(player, profession, recipeName)
  self:EnsureEntryExists(player, profession, recipeName)
  local cache = self:GetCache()
  return cache[player][profession][recipeName]
end

function GuildArtisanCache:Set(player, profession, recipeName, data)
  self:EnsureEntryExists(player, profession, recipeName)
  local cache = self:GetCache()
  cache[player][profession][recipeName] = data
  AceEvent:SendMessage("GUILD_ARTISAN_CACHE_UPDATED")
end

function GuildArtisanCache:InitializePlayer(player)
  local cache = self:GetCache()
  cache[player] = {}
end

function GuildArtisanCache:InitializeProfession(player, profession)
  local cache = self:GetCache()
  cache[player][profession] = {}
end

function GuildArtisanCache:InitializeRecipe(player, profession, recipeName)
  local cache = self:GetCache()
  cache[player][profession][recipeName] = {version = 0, recipeId = 0, level = 0, currentXp = 0, requiredXp = 0}
end

function GuildArtisanCache:SetRecipe(player, profession, recipeId, recipeName, level, currentXp, requiredXp)
  local data = self:Get(player, profession, recipeName)
  if level == nil then level = 0 end
  if currentXp == nil then currentXp = 0 end
  if requiredXp == nil then requiredXp = 0 end
  if data.level == level and data.currentXp == currentXp and data.requiredXp == requiredXp then return false end
  self:Set(player, profession, recipeName, {version = data.version + 1, recipeId = recipeId, level = level, currentXp = currentXp, requiredXp = requiredXp})
  self:Updated(player, profession, recipeName)
end

function GuildArtisanCache:Update(player, profession, recipeId, recipeName, version, level, currentXp, requiredXp)
  local data = self:Get(player, profession, recipeName)
  if version > data.version then
    if player == self.playerName and data.version > 0 then return end -- only update self if we have a freshly-created cache entry
    if level == nil then level = 0 end
    if currentXp == nil then currentXp = 0 end
    if requiredXp == nil then requiredXp = 0 end
    self:Set(player, profession, recipeName, {version = version, recipeId = recipeId, level = level, currentXp = currentXp, requiredXp = requiredXp})
  end
end

function GuildArtisanCache:Updated(player, profession, recipeName)
  self:ClearUpdatedTicker()
  table.insert(self.updated, {player = player, profession = profession, recipeName = recipeName})
  self.updatedTicker = C_Timer.NewTicker(3, function()
    self:ClearUpdatedTicker()
    if #self.updated then
      for i = 1, #self.updated do
        local update = self.updated[i]
        AceEvent:SendMessage("GUILD_ARTISAN_CACHE_UPDATE", update.player, update.profession, update.recipeName)
      end
      self.updated = {}
    end
  end, 0)
end

function GuildArtisanCache:ClearUpdatedTicker()
  if self.updatedTicker then self.updatedTicker:Cancel() end
  self.updatedTicker = nil
end

function GuildArtisanCache:ToMessages()
  local messages = {}
  for player, professions in pairs(self:GetCache()) do
    for profession, recipeNames in pairs(professions) do
      for recipeName, recipe in pairs(recipeNames) do
        local message = self:ToMessage(player, profession, recipeName)
        table.insert(messages, message)
      end
    end
  end
  return messages
end

function GuildArtisanCache:ToMessage(player, profession, recipeName)
  local data = self:Get(player, profession, recipeName)
  return ("%s:%d:%s:%d:%s:%d:%d:%d"):format(player, data.version, profession, data.recipeId, recipeName, data.level, data.currentXp, data.requiredXp)
end

function GuildArtisanCache:GetPlayers()
  local players = {}
  for player, professions in pairs(self:GetCache()) do
    table.insert(players, player)
  end
  return players
end

function GuildArtisanCache:GetProfessions()
  local results = {}
  local seen = {}
  for player, professions in pairs(self:GetCache()) do
    for profession, recipeNames in pairs(professions) do
      if not seen[profession] then
        seen[profession] = true
        table.insert(results, profession)
      end
    end
  end
  return results
end

function GuildArtisanCache:GetRecipesForProfession(profession)
  local results = {}
  local cache = self:GetCache()
  for _, player in pairs(self:GetPlayers()) do
    if cache[player][profession] then
      for name, recipe in pairs(cache[player][profession]) do
        recipe.name = name
        table.insert(results, recipe)
      end
    end
  end
  return results
end

function GuildArtisanCache:GetPlayersForRecipe(recipe)
  local results = {}
  local players = {}
  for player, professions in pairs(self:GetCache()) do
    for profession, recipeNames in pairs(professions) do
      for recipeName, data in pairs(recipeNames) do
        if recipe.name == recipeName and not players[player] then
          players[player] = true
          table.insert(results, {name = player, recipe = recipe})
        end
      end
    end
  end
  return results
end
